from os import getcwd
import git
import argparse
from rich import print
from rich.table import Table
from bossman import Bossman

def init(subparsers: argparse._SubParsersAction):
  parser = subparsers.add_parser("status", help="show resource status")
  parser.add_argument("glob", nargs="?", default="*", help="select resources by glob pattern")
  parser.set_defaults(func=exec)

def exec(bossman: Bossman, glob, *args, **kwargs):
  resources = bossman.get_resources(glob=glob)
  if len(resources):
    table = Table()
    table.add_column("Resource")
    table.add_column("Revision")
    table.add_column("Pending")
    table.add_column("Remote")
    for resource in resources:
      status = bossman.get_resource_status(resource)
      table.add_row(
        resource,
        render_revision(resource, status),
        render_pending(resource, status),
        render_remote(resource, status)
      )
    print(table)
  else:
    print("No resources to show: check the glob pattern if provided, or the configuration.")

def render_revision(resource, status):
  if status.remote_rev and status.remote_rev.local_rev:
    return "[green]{rev} -> v{remote_rev}[/green]".format(rev=status.remote_rev.local_rev, remote_rev=status.remote_rev.remote_rev)
  return "[magenta]{rev}[/magenta]".format(rev=status.local_rev)

def render_pending(resource, status):
  missing_revisions = status.missing_revisions
  if len(missing_revisions) > 0:
    if len(missing_revisions) > 1:
      first, last = status.remote_rev.local_rev, missing_revisions[-1]
      return "{n} ({first}...{last})".format(n=len(missing_revisions), first=first, last=last.id)
    return "{n} ({rev})".format(n=len(missing_revisions), rev=missing_revisions[0].id)
  return "[green]0[/green] :heavy_check_mark:"

def render_remote(resource, status):
  if status.dirty:
    return ":stop_sign: [red][b]dirty[/b][/red]"
  return ":thumbs_up:"


from os.path import abspath, commonpath, join
from fnmatch import fnmatch

from bossman.errors import BossmanError
from bossman.resources import ResourceManager, ResourceStatus
from bossman.abc.resource_type import ResourceTypeABC
from bossman.abc.resource import ResourceABC
from bossman.config import Config, ResourceTypeConfig
from bossman.logging import get_class_logger
from bossman.repo import Repo, Revision, Change


class Bossman:
  def __init__(self, repo_path, config: Config):
    self.config = config
    try:
      self.repo = Repo(repo_path)
    except:
      raise BossmanError("An error occurred, please check that this is a git repository or working tree.")
    self.resource_manager = ResourceManager()
    for config in self.config.resource_types:
      resource_type = ResourceTypeABC.create(config)
      self.resource_manager.register_resource_type(resource_type)
    self.logger = get_class_logger(self)

  def get_resources(self, rev: str = "HEAD", glob: str = "*") -> list:
    resources = self.resource_manager.get_resources(self.repo, rev)
    return list(filter(lambda resource: fnmatch(resource.path, glob), resources))

  def get_resource_status(self, resource: ResourceABC) -> ResourceStatus:
    resource_type = self.resource_manager.get_resource_type(resource.path)
    last_rev = self.repo.get_last_revision(resource.paths)
    remote_rev = resource_type.get_remote_rev(resource)
    missing_revisions = self.repo.get_revisions(remote_rev.local_rev, last_rev.id, resource.paths)
    dirty = resource_type.is_dirty(resource)
    return ResourceStatus(
      local_rev=last_rev.id,
      remote_rev=remote_rev,
      dirty=dirty,
      missing_revisions=list(reversed(missing_revisions))
    )

  def get_revisions(self, since_rev: str = None, until_rev: str = "HEAD", resources: list = None) -> str:
    return self.repo.get_revisions(since_rev, until_rev, resources)

  def apply_change(self, resource: ResourceABC, revision: Revision):
    previous_revision = self.repo.get_last_revision(resource.paths, revision.parent_id)
    resource_type = self.resource_manager.get_resource_type(resource.path)
    resource_type.apply_change(resource, revision, previous_revision)

  def validate(self, resource: ResourceABC):
    resource_type = self.resource_manager.get_resource_type(resource.path)
    resource_type.validate_working_tree(resource)


from abc import ABC, abstractmethod, abstractproperty
import git
from os.path import basename
from bossman.errors import BossmanError
from bossman.logging import get_class_logger
from datetime import datetime
import gitdb.exc

def true(*args, **kwargs):
  return True

def short_rev(rev):
  return rev[:8] if rev else rev

class RepoError(RuntimeError):
  pass

class TreeVisitor:
  """
  Callable to visit all the blobs in a tree.
  """
  def __init__(self, func):
    self.func = func

  def __call__(self, tree: git.objects.tree.Tree):
    def visit(blob, visited):
      if blob.path in visited:
        return # avoid infinite recursion
      visited.add(blob.path)
      if type(blob) == git.Tree:
        for child in blob:
          visit(child, visited)
      elif type(blob) == git.Blob:
        self.func(blob)
    visit(tree, set())

class Change(ABC):
  @staticmethod
  def from_diff(diff: git.diff.Diff):
    if diff.change_type == "R":
      return RenameChange(diff)
    if diff.change_type == "A":
      return AddChange(diff)
    if diff.change_type == "M":
      return ModifyChange(diff)
    if diff.change_type == "D":
      return DeleteChange(diff)
    raise RepoError("unknown change type {}".format(diff.change_type))

  def __init__(self, diff: git.Diff):
    self.diff = diff

  @property
  def path(self):
    return self.diff.b_path

  @property
  def contents(self):
    return self.diff.b_blob.data_stream.read()

  @property
  def diff_text(self):
    return str(self.diff)

  def __str__(self):
    return "{} {}".format(self.diff.change_type, self.path)

  def __rich_console__(self, console, options):
    color = "yellow"
    yield "[{color}]{change_type} {name}[{color}]".format(
      color=color,
      change_type=self.diff.change_type,
      name=basename(self.path),
    )

class AddChange(Change):
  pass

class RenameChange(Change):
  @property
  def from_path(self):
    return self.diff.a_path

class ModifyChange(Change):
  pass

class DeleteChange(Change):
  @property
  def path(self):
    return self.diff.a_path

  @property
  def contents(self):
    return self.diff.a_blob.data_stream.read()



class Revision:
  def __init__(self, commit: git.Commit, diffs: git.Diff):
    self.commit = commit
    self.changes = dict(
      (diff.b_path or diff.a_path, Change.from_diff(diff))
      for diff
      in diffs
    )

  @property
  def id(self):
    return short_rev(self.commit.hexsha)

  @property
  def parent_id(self):
    try:
      return short_rev(self.commit.parents[0].hexsha)
    except IndexError:
      return None

  @property
  def date(self) -> datetime:
    return self.commit.committed_datetime

  @property
  def message(self) -> str:
    return self.commit.message

  @property
  def short_message(self) -> str:
    return self.message.split("\n").pop(0)

  @property
  def author_name(self) -> str:
    return self.commit.author.name

  @property
  def author_email(self) -> str:
    return self.commit.author.email

  def affected_paths(self) -> list:
    return self.changes.keys()

  def show_path(self, path) -> dict:
    contents = self.show_paths([path])
    return contents.get(path, None)

  def show_paths(self, paths: list) -> dict:
    contents = dict()
    def get_blob_contents(blob):
      if blob.path in paths:
        contents[blob.path] = blob.data_stream.read()
    visitor = TreeVisitor(get_blob_contents)
    visitor(self.commit.tree)
    return contents

  def get_changes(self, paths: str) -> Change:
    return list(change for (path, change) in self.changes.items() if path in paths)

  def get_change(self, path: str) -> Change:
    return self.changes.get(path, None)

  def __str__(self):
    s = "[{}] {} {} | {}".format(self.id, self.short_message, self.author_name, self.date)
    if len(self.changes):
      s += "\n\n  "
      s += "\n  ".join(str(change) for change in self.changes.values())
      s += "\n"
    return s

  def __rich_console__(self, console, options):
    yield "[b][{}][/b] {} | {} {}".format(self.id, self.short_message, self.author_name, self.date)
    # for change in self.changes.values():
    #   yield change

class Repo:
  def __init__(self, root):
    self.logger = get_class_logger(self)
    self._repo = git.Repo(root)

  def get_paths(self, rev: str = "HEAD", predicate = true) -> list:
    """
    Lists the paths versioned for revision {rev}, optionally filtered
    by {predicate}.
    """
    try:
      commit = self._repo.rev_parse(rev)
    except gitdb.exc.BadName:
      raise BossmanError("This branch likely has no commits yet.")

    paths = []
    visitor = TreeVisitor(lambda blob: paths.append(blob.path))
    visitor(commit.tree)
    return paths

  def get_last_revision(self, paths: list, rev: str = "HEAD") -> Revision:
    try:
      commit = next(self._repo.iter_commits(rev, paths=paths))
      prev = git.NULL_TREE
      if commit.parents:
        prev = commit.parents[0]
      diffs = commit.diff(prev, paths=paths, R=True)
      return Revision(commit, diffs)
    except StopIteration:
      return None

  def get_revisions(self, since_rev: str = None, until_rev: str = "HEAD",  paths: list = None) -> list:
    commitRange = ("{since_rev}..{until_rev}".format(since_rev=since_rev, until_rev=until_rev)
      if since_rev
      else until_rev)
    commits = self._repo.iter_commits(commitRange, paths=paths)
    revisions = []
    for commit in commits:
      prev = git.NULL_TREE
      if commit.parents:
        prev = commit.parents[0]
      diffs = commit.diff(prev, paths=paths, R=True) # R=True -> reverse
      if len(diffs):
        revisions.append(Revision(commit, diffs))
    return revisions




class BossmanError(RuntimeError):
  """
  Base class for bossman-specific runtime errors.
  """
  def __rich__(self):
    return "{}: {}".format(self.__class__.__name__, " ".join(self.args))

class BossmanConfigurationError(BossmanError):
  """
  Configuration errors indicate that improper options have been passed to bossman.
  """

class MultipleMatchingPluginsError(BossmanError):
  """
  Bossman matches plugins to resources using glob patterns.
  """
  def __init__(self, resource, plugins):
    super(MultipleMatchingPluginsError, self).__init__("Only one plugin should match a resource.")
    self.resource = resource
    self.plugins = plugins



import logging

_console_formatter = logging.Formatter("%(message)s")

_console_handler = logging.StreamHandler()
_console_handler.setFormatter(_console_formatter)

logger = logging.getLogger()

def get_class_logger(instance):
  return logger.getChild(
    "{module_name}.{class_name}".format(
      module_name=instance.__class__.__module__,
      class_name=instance.__class__.__name__
    )
  )


from setuptools import setup, find_namespace_packages

with open("README.md", "r") as fh:
  long_description = fh.read()

with open("VERSION", "r") as fh:
  version = fh.read()

with open("requirements.txt", "r") as fh:
  requirements = fh.readlines()

setup(
  name="bossman",
  version=version,
  author="Anthony Hogg",
  author_email="anthony@hogg.fr",
  description="Automation framework",
  long_description=long_description,
  long_description_content_type="text/markdown",
  url="https://github.com/ynohat/bossman",
  entry_points={
    'console_scripts': [
      'bossman = bossman.cli:main',
    ],
  },
  classifiers=[
    "Programming Language :: Python :: 3",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
  ],
  packages=find_namespace_packages(include=['bossman.*']),
  package_data={
    # 'qgrepper': ['etc/*.xml', 'etc/*.conf', 'templates/*.j2'],
    # 'examples': ['*']
  },
  install_requires=requirements,
  python_requires='>=3.5',
  zip_safe=False
)



from os import getcwd
import git
import argparse
from bossman import Bossman
from bossman.resources import ResourceABC
from rich.console import Console

console = Console()

def init(subparsers: argparse._SubParsersAction):
  parser = subparsers.add_parser("log", help="show resource change history")
  parser.set_defaults(func=exec)

def exec(bossman: Bossman, *args, **kwargs):
  resources = bossman.get_resources()
  revisions = bossman.get_revisions(resources=resources)
  for revision in revisions:
    console.print(revision)
    for resource in resources:
      changes = revision.get_changes(resource.paths)
      if len(changes):
        console.print(resource)
        console.print(*changes)       



from abc import ABC, abstractmethod
from bossman.config import ResourceTypeConfig
from bossman.repo import Revision
from bossman.logging import logger
from bossman.abc.resource import ResourceABC
import parse
from os.path import relpath, join

class ResourceTypeABC(ABC):
  @staticmethod
  def create(config: ResourceTypeConfig):
    import importlib
    plugin = importlib.import_module(config.module)
    return plugin.ResourceType(config)

  """
  Abstract class for resource_types.
  """
  def __init__(self, config: ResourceTypeConfig):
    self.config = config
    self.logger = logger.getChild(
      "{module_name}.{class_name}".format(
        module_name=self.__class__.__module__,
        class_name=self.__class__.__name__
      )
    )
    self.logger.info("config={config}".format(config=str(config)))

  def get_resources(self, paths: list):
    """
    Given a list of paths relative to the repo, determine the subset managed
    by this resource type and return the list of {ResourceABC}s.
    """
    return set(
      resource
      for resource
      in (
        self.get_resource(path)
        for path
        in paths
      )
      if resource != None
    )

  def match(self, path: str) -> bool:
    result = parse.search(self.config.pattern, path)
    if result:
      canonical = self.config.pattern.format(**result.named)
      return path.startswith(canonical)
    return False

  def get_resource(self, path: str) -> ResourceABC:
    result = parse.search(self.config.pattern, path)
    if result:
      canonical = self.config.pattern.format(**result.named)
      if path.startswith(canonical):
        return self.create_resource(canonical, **result.named)
    return None

  @abstractmethod
  def create_resource(self, path, **kwargs) -> ResourceABC:
    pass

  @abstractmethod
  def get_remote_rev(self, resource: ResourceABC) -> str:
    pass

  @abstractmethod
  def is_dirty(self, resource: ResourceABC) -> bool:
    pass

  @abstractmethod
  def apply_change(self, resource: ResourceABC, revision: Revision, previous_revision: Revision):
    pass

  @abstractmethod
  def validate_working_tree(self, resource: ResourceABC):
    pass